"""For backward compatibility only."""
# pylint: skip-file

from neurom.features.bifurcation import *  # pragma: no cover
from neurom.utils import deprecated_module  # pragma: no cover

deprecated_module('Module `neurom.features.bifurcationfunc` is deprecated. Use'
                  '`neurom.features.bifurcation` instead.')  # pragma: no cover
